"""Command-line tool to configure keyboard shortcuts for actions."""

from functools import partial
from itertools import count
from os import getenv
from sys import stderr
from typing import Dict, List, Optional
import re

import gi
gi.require_version('Gio', '2.0')
from gi.repository import Gio

_gsettings_media_keys_key = 'org.gnome.settings-daemon.plugins.media-keys'
_gsettings_custom_keybinding_key = 'org.gnome.settings-daemon.plugins.media-keys.custom-keybinding'
_gsettings_custom_keybinding_path_template = \
    '/org/gnome/settings-daemon/plugins/media-keys/custom-keybindings/custom{}/'
_gsettings_custom_keybinding_path_regex = \
    re.compile(r'/org/gnome/settings-daemon/plugins/media-keys/custom-keybindings/custom([0-9]+)/')
_gsettings_shortcut_name = 'Toggle Terminalle'
_gsettings_shortcut_command_template = \
    'dbus-send --session --type=method_call --dest=party.will.Terminalle' \
        ' /party/will/Terminalle party.will.Terminalle.{}'

def keybind_gsettings(toggle: Optional[List[str]], quit: Optional[List[str]]):
    media_keys = Gio.Settings.new(_gsettings_media_keys_key)
    existing_keybindings = media_keys.get_strv('custom-keybindings')
    new_keybindings = existing_keybindings.copy()

    used_indices = set()
    for path in existing_keybindings:
        match = _gsettings_custom_keybinding_path_regex.match(path)
        if match is not None:
            used_indices.add(int(match[1]))
    indices = count()

    for new_binding, action in _compile_actions(Toggle=toggle, Quit=quit):
        command = _gsettings_shortcut_command_template.format(action)
        # Check if there's a pre-existing binding for the command.
        for path in existing_keybindings:
            custom = Gio.Settings.new_with_path(_gsettings_custom_keybinding_key, path)
            if custom.get_string('command') == command:
                name = custom.get_string('name')
                existing_binding = custom.get_string('binding')
                print(
                    f'Shortcut already exists for {action}: \'{name}\' ({existing_binding})',
                    file=stderr,
                )
                break
        # If not, add the new keybinding.
        else:
            next_index = next(indices)
            while next_index in used_indices:
                next_index = next(indices)
            path = _gsettings_custom_keybinding_path_template.format(next_index)
            new_keybindings.append(path)
            custom = Gio.Settings.new_with_path(_gsettings_custom_keybinding_key, path)
            custom.set_string('name', f'{action} Terminalle')
            custom.set_string('command', command)
            custom.set_string('binding', new_binding)
            print(f'Creating shortcut for {action} ({new_binding})', file=stderr)

    media_keys.set_strv('custom-keybindings', new_keybindings)
    Gio.Settings.sync()

def no_keybind_gsettings():
    media_keys = Gio.Settings.new(_gsettings_media_keys_key)
    existing_keybindings = media_keys.get_strv('custom-keybindings')
    new_keybindings = []

    actions = [
        (action, _gsettings_shortcut_command_template.format(action))
        for action in ['Toggle', 'Quit']
    ]
    for path in existing_keybindings:
        for action, command in actions:
            custom = Gio.Settings.new_with_path(_gsettings_custom_keybinding_key, path)
            if custom.get_string('command') == command:
                binding = custom.get_string('binding')
                custom.set_string('name', '')
                custom.set_string('command', '')
                custom.set_string('binding', '')
                print(f'Removing shortcut for {action} ({binding})', file=stderr)
                break
        else:
            new_keybindings.append(path)

    media_keys.set_strv('custom-keybindings', new_keybindings)
    Gio.Settings.sync()

def keybind_kwriteconfig(toggle: Optional[List[str]], quit: Optional[List[str]]):
    raise NotImplementedError('kwriteconfig is not yet supported')

def no_keybind_kwriteconfig():
    raise NotImplementedError('kwriteconfig is not yet supported')

def keybind_autodetect(toggle: Optional[List[str]], quit: Optional[List[str]]):
    autodetect(
        partial(keybind_gsettings, toggle, quit),
        partial(keybind_kwriteconfig, toggle, quit),
    )

def no_keybind_autodetect():
    autodetect(no_keybind_gsettings, no_keybind_kwriteconfig)

def autodetect(gsettings, kwriteconfig):
    environment = getenv('XDG_CURRENT_DESKTOP')
    if environment == 'GNOME' or environment == 'Unity' or environment == 'X-Cinnamon':
        gsettings()
    elif environment == 'KDE':
        kwriteconfig()
    else:
        reason = 'XDG_CURRENT_DESKTOP is unset' if environment is None else f'\'{environment}\''
        raise RuntimeError(f'Unknown desktop environment: {reason}')

def _compile_actions(**kwargs) -> Dict[str, str]:
    binding_action_list = [
        (binding, action)
        for action, bindings in kwargs.items()
        if bindings is not None
        for binding in bindings
    ]
    binding_action_dict = dict(binding_action_list)
    if len(binding_action_list) > len(binding_action_dict):
        raise ValueError('Cannot use the same binding for multiple actions')
    return binding_action_list



#!/usr/bin/env python3

from os.path import join as join_path
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

from . import Terminalle, load_settings, __version__
from .auto import auto, no_auto, xdg_config_home_path
from .key import (
    keybind_autodetect, keybind_gsettings, keybind_kwriteconfig,
    no_keybind_autodetect, no_keybind_gsettings, no_keybind_kwriteconfig,
)

def build_argparse() -> ArgumentParser:
    parser = ArgumentParser(
        description='A fancy "drop-down" terminal emulateur.',
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument('-v', '--version', action='version', version=__version__)
    parser.add_argument(
        '-c',
        '--config',
        metavar='PATH',
        help='load config settings from PATH',
        default=join_path(xdg_config_home_path, 'terminalle.yaml'),
    )
    parser.add_argument(
        '-s',
        '--show',
        action='store_true',
        help='show the window immediately on startup',
    )

    subparsers = parser.add_subparsers(
        title='subcommands',
        dest='subcommand',
        description='Use `<subcommand> --help` for more information.',
    )
    auto_parser = subparsers.add_parser(
        'auto',
        help='configure auto-start and/or auto-restart',
        description='Install an XDG desktop file to auto-start Terminalle (in hidden mode) on login,'
            ' as well as a DBUS service file to auto-restart Terminalle'
            ' if it\'s not already running when the user attempts to toggle it.',
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    no_auto_parser = subparsers.add_parser(
        'no-auto',
        help='remove auto-(re)start functionality',
        description='Remove the DBUS service file and XDG desktop file'
            ' that were installed using the `auto` subcommand.',
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    key_parser = subparsers.add_parser(
        'key',
        help='configure keyboard shortcuts',
        description='Set up keyboard shortcut(s) to invoke actions (Toggle or Quit).'
            ' Supports GNOME-based desktops (including Unity, Cinnamon, etc.), or KDE.'
            ' The running desktop environment is autodetected by default.',
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    no_key_parser = subparsers.add_parser(
        'no-key',
        help='remove keyboard shortcuts',
        description='Remove any keyboard shortcut(s) to invoke actions (Toggle or Quit).'
            ' Supports GNOME-based desktops (including Unity, Cinnamon, etc.), or KDE.'
            ' The running desktop environment is autodetected by default.',
        formatter_class=ArgumentDefaultsHelpFormatter,
    )

    for sub_parser in [auto_parser, no_auto_parser]:
        target_group = sub_parser.add_mutually_exclusive_group()
        target_group.add_argument(
            '-u',
            '--user',
            action='store_true',
            help='apply to the current user only',
            default=True,
        )
        target_group.add_argument(
            '-s',
            '--system',
            action='store_true',
            help='apply to all users on the system',
            default=False,
        )
        sub_parser.add_argument(
            '-f',
            '--force',
            action='store_true',
            help='modify files even if existing contents are unexpected',
        )
    auto_parser.add_argument(
        '--no-start-on-login',
        action='store_true',
        help='do *not* install XDG desktop file',
    )
    auto_parser.add_argument(
        '--no-restart-if-closed',
        action='store_true',
        help='do *not* install DBUS service file',
    )

    key_parser.add_argument(
        '-t',
        '--toggle',
        action='append',
        metavar='KEYS',
        help='shortcut to toggle window visibility',
    )
    key_parser.add_argument(
        '-q',
        '--quit',
        action='append',
        metavar='KEYS',
        help='shortcut to shut down the application',
    )
    for sub_parser in [key_parser, no_key_parser]:
        desktop = sub_parser.add_mutually_exclusive_group()
        desktop.add_argument(
            '--gsettings',
            action='store_true',
            help='assume the desktop environment is GNOME-based',
        )
        desktop.add_argument(
            '--kwriteconfig',
            action='store_true',
            help='assume the desktop environment is KDE',
        )

    return parser

def main():
    args = build_argparse().parse_args()
    if args.subcommand is None:
        Terminalle(settings=load_settings(args.config), show=args.show).run()
    elif args.subcommand == 'auto':
        auto(args.system, args.force, not args.no_start_on_login, not args.no_restart_if_closed)
    elif args.subcommand == 'no-auto':
        no_auto(args.system, args.force)
    elif args.subcommand == 'key':
        if args.gsettings:
            keybind = keybind_gsettings
        elif args.kwriteconfig:
            keybind = keybind_kwriteconfig
        else:
            keybind = keybind_autodetect
        keybind(args.toggle, args.quit)
    elif args.subcommand == 'no-key':
        if args.gsettings:
            no_keybind = no_keybind_gsettings
        elif args.kwriteconfig:
            no_keybind = no_keybind_kwriteconfig
        else:
            no_keybind = no_keybind_autodetect
        no_keybind()
    else:
        raise ValueError(f'Unexpected subcommand \'{args.subcommand}\'.')

if __name__ == '__main__':
    main()



"""Terminalle application logic."""

from subprocess import Popen
from typing import Callable, Dict, Tuple, Optional

import gi
gi.require_version('Gtk', '4.0')
gi.require_version('Vte', '3.91')
from gi.repository import Gtk, GLib, Gio, Vte

SERVICE_NAME = 'party.will.Terminalle'
OBJECT_PATH = '/party/will/Terminalle'
SERVICE_XML = f'''
<!DOCTYPE node PUBLIC
    "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN"
    "http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">
<node>
  <interface name="{SERVICE_NAME}">
    <method name="Toggle" />
    <method name="Quit" />
  </interface>
</node>
'''

# key names: https://cgit.freedesktop.org/xorg/proto/x11proto/plain/keysymdef.h
# tmux keybinding commands: https://github.com/tmux/tmux/blob/3.5a/key-bindings.c#L347
_tmux_mode_commands = [
    ('exclam', ['tmux', 'break-pane']),
    ('quotedbl', ['tmux', 'split-window']),
    ('numbersign', ['tmux', 'list-buffers']),
    ('dollar', ['tmux', 'command-prompt', '-I', '#S', 'rename-session -- \'%%\'']),
    ('percent', ['tmux', 'split-window', '-h']),
    ('ampersand', ['tmux', 'confirm-before', '-p', 'kill-window #W? (y/n)', 'kill-window']),
    ('apostrophe', ['tmux', 'command-prompt', '-T', 'window-target', '-pindex', 'select-window -t \':%%\'']),
    ('parenleft', ['tmux', 'switch-client', '-p']),
    ('parenright', ['tmux', 'switch-client', '-n']),
    ('comma', ['tmux', 'command-prompt', '-I', '#W', 'rename-window -- \'%%\'']),
    ('minus', ['tmux', 'delete-buffer']),
    ('period', ['tmux', 'command-prompt', '-T', 'target', 'move-window -t \'%%\'']),
    ('colon', ['tmux', 'command-prompt']),
    ('semicolon', ['tmux', 'last-pane']),
    ('equal', ['tmux', 'choose-buffer', '-Z']),
    ('bracketleft', ['tmux', 'copy-mode']),
    ('bracketright', ['tmux', 'paste-buffer']),
    ('braceleft', ['tmux', 'swap-pane', '-U']),
    ('braceright', ['tmux', 'swap-pane', '-D']),
]

class Terminalle:
    """Manages the D-Bus service and the terminal window."""

    def __init__(self, settings: Dict[str, object], show: bool):
        """Initialize the window and VTE widget."""
        self.app = Gtk.Application(application_id=SERVICE_NAME)
        self.app.connect('activate', self._on_activate)
        self.settings = settings
        self.show_on_startup = show

    def _on_activate(self, app: Gtk.Application):
        """Create and show the main window."""
        window = Gtk.ApplicationWindow(application=app)
        window.set_title('Terminalle')
        window.set_icon_name('utilities-terminal')
        window.set_decorated(False)
        # Make the application window's background transparent using CSS.
        css_provider = Gtk.CssProvider()
        css_provider.load_from_data(b'* { background-color: transparent; }')
        Gtk.StyleContext.add_provider_for_display(
            window.get_display(),
            css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
        )
        # Maximize the window because fullscreen does not support transparency.
        # https://gitlab.freedesktop.org/wayland/wayland-protocols/-/issues/116
        window.maximize()
        if self.settings['autohide']:
            focus_controller = Gtk.EventControllerFocus()
            focus_controller.connect('leave', self._autohide)
            window.add_controller(focus_controller)

        terminal = Vte.Terminal()
        terminal.set_font(font_desc=self.settings['font'])
        terminal.set_allow_bold(True)
        terminal.set_allow_hyperlink(True)
        bg = self.settings['colors'][0].copy()
        bg.alpha = self.settings['opacity']
        terminal.set_colors(background=bg, palette=self.settings['colors'])
        terminal.connect('child-exited', self._term_exited)
        window.set_child(terminal)

        self.window = window
        self.terminal = terminal

        terminal.spawn_async(
            Vte.PtyFlags.DEFAULT,             # PTY flags
            self.settings['home'],            # working directory
            [self.settings['shell']],         # command-line arguments
            None,                             # environment variables
            GLib.SpawnFlags.DEFAULT,          # spawn flags
            None,                             # child setup callback
            (),                               # child setup callback arguments
            -1,                               # timeout
            None,                             # cancellable
            self._term_spawn_async_callback,  # spawn callback
            (),                               # spawn callback arguments
        )

        # Set up keyboard shortcuts.
        shortcut_controller = Gtk.ShortcutController()
        shortcut_controller.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        _add_shortcut(shortcut_controller, '<Control><Shift>c', self._copy_clipboard)
        _add_shortcut(shortcut_controller, '<Control><Shift>v', self._paste_clipboard)
        if self.settings['tmux']:
            for key_name, cmd in _tmux_mode_commands:
                _add_shortcut(shortcut_controller, f'<Control>{key_name}', _run_cmd_handler(cmd))
        window.add_controller(shortcut_controller)

    def _term_spawn_async_callback(
        self, terminal: Vte.Terminal, pid: int, error: Optional[GLib.Error]
    ):
        """Finish starting up after the terminal has been spawned."""
        if error is not None:
            self.quit()
            raise RuntimeError(
                f'Error spawning VTE [{error.domain}:{error.code}]: {error.message}'
            )
        if self.show_on_startup:
            self.window.present()
            self.window.grab_focus()

        self.methods = {
            'Toggle': self.toggle,
            'Quit': self.quit,
        }
        service = Gio.DBusNodeInfo.new_for_xml(SERVICE_XML)
        self.app.get_dbus_connection().register_object(
            OBJECT_PATH, service.interfaces[0], self._on_method_call
        )

    def _on_method_call(
        self,
        connection: Gio.DBusConnection,
        sender: str,
        object_path: str,
        interface_name: str,
        method_name: str,
        parameters: Tuple,
        invocation: Gio.DBusMethodInvocation,
    ):
        """Handle a D-Bus method invocation."""
        self.methods[method_name]()
        invocation.return_value()

    def run(self):
        """Run the GTK application."""
        self.app.run(None)

    def toggle(self):
        """Toggle window visibility."""
        GLib.idle_add(self._toggle)

    def _toggle(self):
        if self.window.is_active():
            self.window.hide()
        else:
            if not self.window.props.visible:
                self.window.present()
            self.window.grab_focus()

    def _copy_clipboard(self, widget: Gtk.Widget, args: Optional[GLib.Variant]) -> bool:
        self.terminal.copy_clipboard_format(Vte.Format.TEXT)
        return True

    def _paste_clipboard(self, widget: Gtk.Widget, args: Optional[GLib.Variant]) -> bool:
        self.terminal.paste_clipboard()
        return True

    def _autohide(self, controller: Gtk.EventControllerFocus):
        """Hide the window when it loses focus."""
        GLib.idle_add(self.window.hide)

    def _term_exited(self, terminal: Vte.Terminal, status: int):
        """Close the window automatically when the terminal exits."""
        self.quit()

    def quit(self):
        """Quit the GTK application."""
        GLib.idle_add(self.app.quit)

def _add_shortcut(
    controller: Gtk.ShortcutController,
    trigger: str,
    handler: Callable[[Gtk.Widget, Optional[GLib.Variant], object], bool],
):
    """Set up a keyboard shortcut."""
    controller.add_shortcut(
        Gtk.Shortcut.new(
            Gtk.ShortcutTrigger.parse_string(trigger),
            Gtk.CallbackAction.new(handler),
        )
    )

def _run_cmd_handler(cmd: str) -> Callable[[Gtk.Widget, Optional[GLib.Variant]], bool]:
    def _handler(widget: Gtk.Widget, args: Optional[GLib.Variant]) -> bool:
        Popen(cmd)
        return True
    return _handler
